/* =====================================
=  Program Name: abundancebin
=  File Name: profileManager.cpp
=  Version: 1.0
=  Developer: Yu-Wei Wu <yuwwu@indiana.edu>, and Yuzhen Ye <yye@indiana.edu>
=  Affiliation: School of Informatics, Indiana University, Bloomington
=
=  abundancebin is free software under the terms of the GNU General Public
=  License as published by the Free Software Foundation.
======================================*/

#include "profileManager.h"
#include <fstream>
#include <math.h>

const int profileManager::PREDEFINED_MAX_GENOME_SIZE = 5000000;
const int profileManager::PREDEFINED_MIN_GENOME_SIZE = 50000;
const double profileManager::CONST_E = (float)2.71828183;
const float profileManager::CONVERGE_THRESHOLD = (float)0.00001;
const float profileManager::SMALL_VALUE = (float)-9999;
const float profileManager::VERY_SMALL_VALUE = (float)-99999999;
const int profileManager::TYPE_INPUT = 1001;
const int profileManager::TYPE_CLASSIFY = 1002;
const int profileManager::TYPE_RECUR = 1003;
const int profileManager::DEBUG_MUST = 2001;
const int profileManager::DEBUG_NORMAL = 2002;

profileManager::profileManager(int input_len)
{
	mem_handler = new myMalloc();
	len_mer = input_len;
	initParam();
}

void profileManager::readFasta(char *inputFile)
{
	if (logfile == NULL)
	{
		logfile = (char*)mem_handler->allocate(strlen(inputFile) + 5);
		sprintf(logfile, "%s.out", inputFile);
	}
	log = new fstream(logfile, ios::out);

	inputFasta = inputFile;
	parseFasta(inputFasta, TYPE_INPUT, 0, 0, 0);
}

profileManager::~profileManager()
{
	delPoissonClass();
	if (prevClassSize != NULL)
	{
		free(prevClassSize);
	}
	if (prevClassAbundance != NULL)
	{
		free(prevClassAbundance);
	}
	if (classSize != NULL)
	{
		free(classSize);
	}
	if (classAbundance != NULL)
	{
		free(classAbundance);
	}
	if (resultSeqNum != NULL)
	{
		free(resultSeqNum);
	}
	if (resultClass != NULL)
	{
		free(resultClass);
	}
	log->close();
	delete log;
	delete mem_handler;
}

void profileManager::setMerLen(int input_len)
{
	len_mer = input_len;
}

void profileManager::setMaxRun(int maxRun)
{
	EM_maxrun = maxRun;
}

void profileManager::setVerbose(bool verbose)
{
	is_verbose = verbose;
}

void profileManager::setOutputFile(char *input)
{
	logfile = input;
}

bool profileManager::parseFasta(char *inputFasta, int type, int exclude, int exclude_max, int part)
{
	int iFlag;
	unsigned int iLen;
	unsigned _int64 tempNum, tempRevNum;
	int i, j;
	char *tempMer, *tempRevMer, *tempHeader;
	fstream fs(inputFasta, ios::in);
	iFlag = 0;
	// For classification
	float maxfloat;
	float *tempfarray, *tempResultArray;
	unsigned int included_mer, total_mer;

	if (type == TYPE_INPUT)
	{
		kmer_highest_count = 0;
	}
	if(fs.is_open() == false)
	{
		is_Fasta = false;
		return false;
	}
	else
	{
		is_Fasta = true;
	}

	tempMer = (char*)mem_handler->allocate(len_mer + 1);
	memset(tempMer, '\0', len_mer + 1);
	tempRevMer = (char*)mem_handler->allocate(len_mer + 1);
	memset(tempRevMer, '\0', len_mer + 1);
	if (type == TYPE_INPUT || type == TYPE_RECUR)
	{
		if (profileHash != NULL)
		{
			delete profileHash;
		}
		profileHash = new myHash(len_mer);
	}
	else if (type == TYPE_CLASSIFY)
	{
		if (resultClass != NULL)
		{
			free(resultClass);
		}
		if (resultSeqNum != NULL)
		{
			free(resultSeqNum);
		}

		resultSeqNum = (int*)mem_handler->allocate((class_num + 1) * sizeof(int));
		memset(resultSeqNum, '\0', (class_num + 1) * sizeof(int));
		resultClass = (int*)mem_handler->allocate(class_num * sizeof(int));
		memset(resultClass, '\0', class_num * sizeof(int));
		for (i = 0; i < class_num; i++)
		{
			resultClass[i] = i;
		}

		tempfarray = (float*)mem_handler->allocate(class_num * sizeof(float));
		tempResultArray = (float*)mem_handler->allocate(class_num * sizeof(float));

		total_likelihood = 0;
	}

	start_time = time(NULL);
	seqnum = 0;
	sprintf(logstr, "Loading sequences...\n");
	writelog(logstr, DEBUG_MUST);
	while (!fs.eof())
	{
		memset(str, '\0', 1000);
		fs.getline(str,1000);
		if (str[0] == '>' || fs.eof())
		{
			seqnum++;
			if (seqnum % 1000 == 0)
			{
				if (type == TYPE_INPUT || type == TYPE_RECUR)
				{
					sprintf(logstr, "Loaded %d sequences, kmer count = %u\n", seqnum, profileHash->reportTotalNum());
					writelog(logstr, DEBUG_NORMAL);
				}
				if (type == TYPE_CLASSIFY)
				{
					sprintf(logstr, "Loaded %d sequences\n", seqnum);
					writelog(logstr, DEBUG_NORMAL);
				}
			}
			if (iFlag == 0)
			{
				iFlag = 1;
				if (type == TYPE_INPUT)
				{
					i = strlen(str);
					if (i > 0)
					{
						tempHeader = (char*)mem_handler->allocate(i);
						memset(tempHeader, '\0', i);
						memcpy(tempHeader, str + 1, i - 1);
					}
				}
				if (type == TYPE_RECUR || type == TYPE_CLASSIFY)
				{
					seqNow = seqStart;
				}
			}
			else
			{
				if (type == TYPE_INPUT)
				{
					// Append seq_lnk
					if (seqStart == NULL)
					{
						seqStart = (seq_lnk*)mem_handler->allocate(sizeof(seq_lnk));
						seqNow = seqStart;
					}
					else
					{
						seqNow->next = (seq_lnk*)mem_handler->allocate(sizeof(seq_lnk));
						seqNow = seqNow->next;
					}
					memset(seqNow, '\0', sizeof(seq_lnk));
					seqNow->header = tempHeader;
				}

				// Analyze current sequence (tempseq);
				iLen = tempseq->length();
				if (type == TYPE_CLASSIFY)
				{
					included_mer = 0;
					total_mer = 0;
					memset(tempfarray, '\0', class_num * sizeof(float));
				}
				if ((type == TYPE_CLASSIFY && seqNow->recur_mark != part) || (type == TYPE_RECUR && seqNow->recur_mark != part))
				{
					seqNow = seqNow->next;
					delete tempseq;
					tempseq = NULL;
					continue;
				}
				for (i = 0; i < (int)iLen - len_mer + 1; i++)
				{
					j = tempseq->copy(tempMer, len_mer, i);
					tempNum = calNum(tempMer);
					if (err != -1)
					{
						if (type == TYPE_INPUT || (type == TYPE_RECUR && seqNow->recur_mark == part))
						{
							profileHash->add(tempNum);
						}
						else if (type == TYPE_CLASSIFY)
						{
							profileRecent = profileHash->find(tempNum);
							total_mer++;
							if (usability[profileRecent->count] == true && profileRecent->count > exclude && (exclude_max == 0 || profileRecent->count <= exclude_max))
							{
								for (j = 0; j < class_num; j++)
								{
									if (count_poisson_class[profileRecent->count][j] == 0)
									{
										tempfarray[j] = tempfarray[j] + SMALL_VALUE;
									}
									else
									{
										tempfarray[j] = tempfarray[j] + log10(count_poisson_class[profileRecent->count][j]);
									}
								}
								included_mer++;
							}
						}
					}
					revSeq(tempMer, tempRevMer);
					tempRevNum = calNum(tempRevMer);
					if (tempRevNum != tempNum && err != -1)
					{
						if (type == TYPE_INPUT || (type == TYPE_RECUR && seqNow->recur_mark == part))
						{
							profileHash->add(tempRevNum);
						}
						else if (type == TYPE_CLASSIFY)
						{
							profileRecent = profileHash->find(tempNum);
							total_mer++;
							if (usability[profileRecent->count] == true && profileRecent->count > exclude && (exclude_max == 0 || profileRecent->count <= exclude_max))
							{
								for (j = 0; j < class_num; j++)
								{
									if (count_poisson_class[profileRecent->count][j] == 0)
									{
										tempfarray[j] = tempfarray[j] + SMALL_VALUE;
									}
									else
									{
										tempfarray[j] = tempfarray[j] + log10(count_poisson_class[profileRecent->count][j]);
									}
								}
								included_mer++;
							}
						}
					}
				}
				if (type == TYPE_CLASSIFY)
				{
					memset(tempResultArray, '\0', class_num * sizeof(float));
					for (i = 0; i < class_num; i++)
					{
						for (j = 0; j < class_num; j++)
						{
							tempResultArray[i] = tempResultArray[i] + (double)pow((double)10, (double)(tempfarray[j] - tempfarray[i]));
						}
						tempResultArray[i] = 1 / tempResultArray[i];
					}

					maxfloat = (float)included_mer / (float)total_mer;
					if (maxfloat <= coverage_threshold)
					{
						seqNow->class_destination = -1;
						resultSeqNum[class_num]++;
					}
					else
					{
						//maxfloat = SMALL_VALUE;
						maxfloat = 0;
						j = -1;
						for (i = 0; i < class_num; i++)
						{
							if (maxfloat < tempResultArray[i])
							{
								maxfloat = tempResultArray[i];
								j = i;
							}
						}

						if (maxfloat <= read_threshold)
						{
							seqNow->class_destination = -1;
							resultSeqNum[class_num]++;
						}
						else
						{
							seqNow->class_destination = j;
							resultSeqNum[j]++;
							total_likelihood = total_likelihood + log10(maxfloat);
						}
					}
				}
				if (type == TYPE_INPUT)
				{
					i = strlen(str);
					if (i > 0)
					{
						tempHeader = (char*)mem_handler->allocate(i);
						memset(tempHeader, '\0', i);
						memcpy(tempHeader, str + 1, i - 1);
					}
				}
				if (type == TYPE_RECUR || type == TYPE_CLASSIFY)
				{
					seqNow = seqNow->next;
				}

				delete tempseq;
				tempseq = NULL;
			}
		}
		else if (str[0] != '\0')
		{
			if (tempseq == NULL)
			{
				tempseq = new string(str);
			}
			else
			{
				tempseq->append(str);
			}
		}
	}

	free(tempMer);
	free(tempRevMer);

	if (type == TYPE_INPUT)
	{
		kmer_highest_count = profileHash->returnHighestCount();
		//sprintf(str, "Total number of kmers: %u\nTotal occupied memory: %lu\n", profileHash->reportTotalNum(), profileHash->reportTotalMem());
		sprintf(logstr, "Loaded sequences: %d\nLoaded kmers: %u\n", seqnum - 1, profileHash->reportTotalNum());
		writelog(logstr, DEBUG_MUST);
	}
	else if (type == TYPE_CLASSIFY)
	{
		free(tempfarray);
		free(tempResultArray);

		writelog("Classified sequences:\n", DEBUG_MUST);
		for (i = 0; i < class_num; i++)
		{
			sprintf(logstr, "\tclass %04d: %d\n", i, resultSeqNum[i]);
			writelog(logstr, DEBUG_MUST);
		}
		sprintf(logstr, "Unclassified sequences: %d\n\n", resultSeqNum[class_num]);
		writelog(logstr, DEBUG_MUST);
	}

	end_time = time(NULL);
	writelog("Elasped time for constructing the profile sequences:\n", DEBUG_MUST);
	logElapsedTime();

	return true;
}

void profileManager::initParam()
{
	seqStart = NULL;
	seqNow = NULL;
	profileHash = NULL;
	tempseq = NULL;
	err = 0;
	classAbundance = NULL;
	logfile = NULL;
	classSize = NULL;
	prevClassAbundance = NULL;
	prevClassSize = NULL;
	resultSeqNum = NULL;
	resultClass = NULL;
	count_poisson_class = NULL;
	usability = NULL;
	is_len = true;
	is_Fasta = false;
	is_verbose = false;
	seqnum = 0;
	kmer_threshold = 0.5;
	read_threshold = 0.5;
	coverage_threshold = (float)0.1;
	abundance_diff_threshold = 0.5;
	size_threshold = 400000;
	min_seq_threshold = (float)0.03;
	EM_maxrun = 100;
	total_likelihood = 0;
	trial_class_num = 0;
}

void profileManager::delPoissonClass()
{
	int i;
	if (count_poisson_class != NULL)
	{
		for (i = 0; i < kmer_highest_count + 1; i++)
		{
			free(count_poisson_class[i]);
		}
		free(count_poisson_class);
	}
	if (usability != NULL)
	{
		free(usability);
	}
}

unsigned _int64 profileManager::calNum(char *str)
{
	int i;
	unsigned _int64 num;
	num = 0;
	err = 0;
	for (i = 0; i < len_mer; i++)
	{
		num = num * 4;
		switch (str[i])
		{
			case 'A':
			case 'a':
				num = num + 0;
			break;

			case 'T':
			case 't':
				num = num + 1;
			break;

			case 'C':
			case 'c':
				num = num + 2;
			break;

			case 'G':
			case 'g':
				num = num + 3;
			break;

			default:
				num = 0;
				err = -1;
			break;
		}
		if (err == -1)
		{
			break;
		}
	}
	return num;
}

void profileManager::initializeEM(int input_classnum)
{
	int i, j;

	class_num = input_classnum;
	kmer_threshold = 1 / (float)class_num;

	writelog("Starting Expectation Maximization process...\n", DEBUG_MUST);
	start_time = time(NULL);

	if (classSize != NULL)
	{
		free(classSize);
	}
	if (classAbundance != NULL)
	{
		free(classAbundance);
	}
	if (prevClassSize != NULL)
	{
		free(prevClassSize);
	}
	if (prevClassAbundance != NULL)
	{
		free(prevClassAbundance);
	}

	// Initialize the poisson distribution parameters
	classSize = (double*)mem_handler->allocate(class_num * sizeof(double));
	memset(classSize, '\0', class_num * sizeof(double));
	classAbundance = (double*)mem_handler->allocate(class_num * sizeof(double));
	memset(classAbundance, '\0', class_num * sizeof(double));
	prevClassSize = (double*)mem_handler->allocate(class_num * sizeof(double));
	memset(prevClassSize, '\0', class_num * sizeof(double));
	prevClassAbundance = (double*)mem_handler->allocate(class_num * sizeof(double));
	memset(prevClassAbundance, '\0', class_num * sizeof(double));

	srand(time(NULL));

	j = 10 * (class_num - 1);
	for (i = 0; i < class_num; i++)
	{
		// Set abundance randomly between 0.1 and 10
		// classAbundance[i] = ((rand() / (float(RAND_MAX) + 1)) * (10 - 0.1)) + 0.1;

		// Set abundance value with exact values
		classAbundance[i] = j;
		j = j - 10;
		if (j == 0)
		{
			j = 1;
		}

		if (is_len == true)
		{
			/*
			// Set genome size randomly between 50K to 5MB
			classSize[i] = rand() % PREDEFINED_MAX_GENOME_SIZE + 1;
			if (classSize[i] < PREDEFINED_MIN_GENOME_SIZE)
			{
				classSize[i] = classSize[i] + PREDEFINED_MIN_GENOME_SIZE;
			}
			*/

			// Set genome size fixed
			classSize[i] = 1000000;

			// Set genome size to normal distribution
			// mean = 2916752.3, variance = 549285058494950, standard deviation = 1835714.3
			/*
			classSize[i] = 0;
			for (k = 0; k < 12; k++)
			{
				classSize[i] = classSize[i] + (rand() / (float)((double)(RAND_MAX) + (double)1));
			}
			classSize[i] = ((classSize[i] / 6) * 1835714) + (2916752 - 6);
			*/
			
		}
		else
		{
			classSize[i] = 0;
		}
	}

	// Initialize the probability array of every counts
	delPoissonClass();
	count_poisson_class = (double**)mem_handler->allocate((kmer_highest_count + 1) * sizeof(double*));
	for (i = 0; i < kmer_highest_count + 1; i++)
	{
		count_poisson_class[i] = (double*)mem_handler->allocate(class_num * sizeof(double));
	}
	usability = (bool*)mem_handler->allocate((kmer_highest_count + 1) * sizeof(bool));
}

void profileManager::performEM(int exclude, int exclude_max)
{
	int i, j, k, cnt;
	int zeroflag, convflag;
	double fTemp;

	k = 0;
	while (1)
	{
		// Expectation
		// For each count except count = 0, calculate the probability that the given profile belongs to each class
		for (cnt = 1; cnt <= kmer_highest_count; cnt++)
		{
			for (i = 0; i < class_num; i++)
			{
				count_poisson_class[cnt][i] = 0;
				if (is_len == true)
				{
					for (j = 0; j < class_num; j++)
					{
						count_poisson_class[cnt][i] = (double)count_poisson_class[cnt][i] + (double)((double)classSize[j] * (double)pow((double)(classAbundance[j] / classAbundance[i]), (double)cnt) * (double)pow(CONST_E, (double)(classAbundance[i] - classAbundance[j])));
					}
					count_poisson_class[cnt][i] = (double)classSize[i] / count_poisson_class[cnt][i];
				}
				else
				{
					for (j = 0; j < class_num; j++)
					{
						count_poisson_class[cnt][i] = (double)count_poisson_class[cnt][i] + (double)((double)pow((double)(classAbundance[j] / classAbundance[i]), (double)cnt) * (double)pow(CONST_E, (double)(classAbundance[i] - classAbundance[j])));
					}
					count_poisson_class[cnt][i] = (double)1 / count_poisson_class[cnt][i];
				}
			}
		}

		// Maximization
		// classSize and classAbundance
		for (i = 0; i < class_num; i++)
		{
			prevClassSize[i] = classSize[i];
			classSize[i] = 0;
			prevClassAbundance[i] = classAbundance[i];
			classAbundance[i] = 0;
		}
		profileHash->resetIndex();
		while ((profileRecent = profileHash->next()) != NULL)
		{
			if (profileRecent->count > exclude && (exclude_max == 0 || profileRecent->count <= exclude_max))
			{
				for (i = 0; i < class_num; i++)
				{
					classSize[i] = classSize[i] + count_poisson_class[profileRecent->count][i];
					classAbundance[i] = classAbundance[i] + (count_poisson_class[profileRecent->count][i] * profileRecent->count);
				}
			}
		}
		for (i = 0; i < class_num; i++)
		{
			classAbundance[i] = classAbundance[i] / classSize[i];
		}

		// Write log
		writelog(k, DEBUG_NORMAL);
		for (i = 0; i < class_num; i++)
		{
			writelog("\t", DEBUG_NORMAL);
			writelog(classAbundance[i], DEBUG_NORMAL);
		}
		if (is_len == true)
		{
			writelog("\t", DEBUG_NORMAL);
			for (i = 0; i < class_num; i++)
			{
				writelog("\t", DEBUG_NORMAL);
				writelog(classSize[i], DEBUG_NORMAL);
			}
		}
		writelog("\n", DEBUG_NORMAL);

		// Check for convergence
		if (k >= EM_maxrun)
		{
			writelog("Exceeding maximum runs. Stop EM process.\n", DEBUG_NORMAL);
			break;
		}

		zeroflag = 0;
		convflag = 0;
		for (i = 0; i < class_num; i++)
		{
			if (is_len == true)
			{
				if (classAbundance[i] == 0 || classSize[i] == 0)
				{
					zeroflag = 1;
				}
				else
				{
					fTemp = (classAbundance[i] - prevClassAbundance[i]) / prevClassAbundance[i];
					if (fTemp < 0)
					{
						fTemp = fTemp * -1;
					}
					if (fTemp < CONVERGE_THRESHOLD)
					{
						convflag++;
					}
					fTemp = (classSize[i] - prevClassSize[i]) / prevClassSize[i];
					if (fTemp < 0)
					{
						fTemp = fTemp * -1;
					}
					if (fTemp < CONVERGE_THRESHOLD)
					{
						convflag++;
					}
				}
			}
			else
			{
				if (classAbundance[i] == 0)
				{
					zeroflag = 1;
				}
				else
				{
					fTemp = (classAbundance[i] - prevClassAbundance[i]) / prevClassAbundance[i];
					if (fTemp < 0)
					{
						fTemp = fTemp * -1;
					}
					if (fTemp < CONVERGE_THRESHOLD)
					{
						convflag = convflag + 2;
					}
				}
			}
		}

		if (is_len == true)
		{
			if (zeroflag == 1 || convflag == class_num * 2)
			{
				break;
			}
		}
		else
		{
			if (zeroflag == 1 || convflag == class_num)
			{
				break;
			}
		}

		k++;
	}
	writelog("Expectation Maximization process ends. Final parameters are:\n", DEBUG_MUST);

	for (i = 0; i < class_num; i++)
	{
		writelog(classAbundance[i], DEBUG_MUST);
		writelog("\t", DEBUG_MUST);
	}
	if (is_len == true)
	{
		writelog("\t", DEBUG_MUST);
		for (i = 0; i < class_num; i++)
		{
			writelog(classSize[i], DEBUG_MUST);
			writelog("\t", DEBUG_MUST);
		}
	}
	writelog("\n", DEBUG_MUST);

	// Check the usability of reads
	for (cnt = 1; cnt <= kmer_highest_count; cnt++)
	{
		usability[cnt] = false;
		for (i = 0; i < class_num; i++)
		{
			if (count_poisson_class[cnt][i] > kmer_threshold)
			{
				usability[cnt] = true;
			}
		}
	}

	end_time = time(NULL);
	writelog("Elasped time for EM process:\n", DEBUG_MUST);
	logElapsedTime();
}

void profileManager::classifySeq(int exclude, int exclude_max, int part)
{
	parseFasta(inputFasta, TYPE_CLASSIFY, exclude, exclude_max, part);
}

void profileManager::writeResult()
{
	int i;
	fstream **fsArray;
	fstream *fsUnclassify;

	fsArray = (fstream**)mem_handler->allocate(class_num * sizeof(fstream*));
	for (i = 0; i < class_num; i++)
	{
		sprintf(str, "%s.%d", logfile, i + 1);
		fsArray[i] = new fstream(str, ios::out);
	}
	sprintf(str, "%s.unclassified", logfile);
	fsUnclassify = new fstream(str, ios::out);

	writelog("Writing classified results into files...\n", DEBUG_MUST);

	seqNow = seqStart;
	while (seqNow != NULL)
	{
		if (seqNow->class_destination == -1)
		{
			sprintf(str, "%s", seqNow->header);
			fsUnclassify->write(str, strlen(str));
			fsUnclassify->write("\n", 1);
		}
		else
		{
			sprintf(str, "%s", seqNow->header);
			fsArray[seqNow->class_destination]->write(str, strlen(str));
			fsArray[seqNow->class_destination]->write("\n", 1);
		}
		seqNow = seqNow->next;
	}

	for (i = 0; i < class_num; i++)
	{
		fsArray[i]->close();
		delete(fsArray[i]);
	}
	delete(fsUnclassify);
}

void profileManager::writeFASTA()
{
	int iFlag;
	int i, loc;
	myStringHash *strhash;

	fstream **fsArray;
	fstream *fsUnclassify;

	fsArray = (fstream**)mem_handler->allocate(class_num * sizeof(fstream*));
	for (i = 0; i < class_num; i++)
	{
		sprintf(str, "%s.%d", logfile, i + 1);
		fsArray[i] = new fstream(str, ios::out);
	}
	sprintf(str, "%s.unclassified", logfile);
	fsUnclassify = new fstream(str, ios::out);

	writelog("Writing classified results with FULL SEQUENCE into files...\n", DEBUG_MUST);

	fstream fs(inputFasta, ios::in);
	iFlag = 0;

	// Use another hash to record the target
	strhash = new myStringHash();
	seqNow = seqStart;
	while (seqNow != NULL)
	{
		strhash->add(seqNow->header, seqNow->class_destination);
		seqNow = seqNow->next;
	}

	if(fs.is_open() == false)
	{
		is_Fasta = false;
	}
	else
	{
		is_Fasta = true;
	}

	seqnum = 0;
	while (!fs.eof())
	{
		memset(str, '\0', 1000);
		fs.getline(str,1000);
		if (str[0] == '>')
		{
			// Find where to write the file
			seqnum++;
			/*
			if (seqnum % 1000 == 0)
			{
				sprintf(logstr, "Written %d sequences\n", seqnum);
				writelog(logstr, DEBUG_NORMAL);
			}
			*/
			loc = strhash->find(str + 1);
			if (loc == -1)
			{
				fsUnclassify->write(str, strlen(str));
				fsUnclassify->write("\n", 1);
			}
			else
			{
				fsArray[loc]->write(str, strlen(str));
				fsArray[loc]->write("\n", 1);
			}
		}
		else if (str[0] != '\0')
		{
			if (loc == -1)
			{
				fsUnclassify->write(str, strlen(str));
				fsUnclassify->write("\n", 1);
			}
			else
			{
				fsArray[loc]->write(str, strlen(str));
				fsArray[loc]->write("\n", 1);
			}
		}
	}

	for (i = 0; i < class_num; i++)
	{
		fsArray[i]->close();
		delete(fsArray[i]);
	}
	fsUnclassify->close();
	delete(fsUnclassify);
}

void profileManager::writeBinFeature()
{
	int i;

	fstream *fsBin;

	sprintf(str, "%s.bin_feature", logfile);
	fsBin = new fstream(str, ios::out);

	sprintf(str, "Total number of bins: %d\n\n", class_num);
	fsBin->write(str, strlen(str));

	for (i = 0; i < class_num; i++)
	{
		sprintf(str, "Bin %d\nGenome Abundance: %0.2f\nGenome Size: %0.2f\n\n", i + 1, classAbundance[i], classSize[i] / 2);
		fsBin->write(str, strlen(str));
	}

	fsBin->close();
	delete(fsBin);
}

void profileManager::writeKmerCount()
{
	unsigned int *count;
	int i;
	fstream *fcount;

	count = (unsigned int*)mem_handler->allocate((kmer_highest_count + 1) * sizeof(unsigned int));
	memset(count, '\0', (kmer_highest_count + 1) * sizeof(unsigned int));

	sprintf(str, "%s.count", logfile);
	fcount = new fstream(str, ios::out);

	profileHash->resetIndex();
	while ((profileRecent = profileHash->next()) != NULL)
	{
		count[profileRecent->count]++;
	}
	for (i = 1; i <= kmer_highest_count; i++)
	{
		sprintf(str, "%d\t%d\n", i, count[i]);
		fcount->write(str, strlen(str));
	}

	fcount->close();
	delete(fcount);
}

int profileManager::binNumberReduction(int exclude, int exclude_max)
{
	int i, j, tmp_classnum;
	bool *remove_indicator;
	bool change;
	// Check from the highest abundance, and merge the bins to the lowest-possible one if the sequence number < threshold
	tmp_classnum = class_num;
	remove_indicator = (bool*)mem_handler->allocate(class_num * sizeof(bool));
	memset(remove_indicator, '\0', class_num * sizeof(bool));
	change = false;
	for (i = 0; i < class_num; i++)
	{
		if (remove_indicator[i] == false && classSize[i] < size_threshold)
		{
			sprintf(logstr, "binNumberReduction: bin %d removed. Length %0.4f too small.\n", i, classSize[i]);
			writelog(logstr, DEBUG_MUST);
			remove_indicator[i] = true;
			tmp_classnum--;
			change = true;
		}
	}
	if (change == false)
	{
		for (i = 0; i < class_num; i++)
		{
			if (i > 0 && ((classAbundance[i] * 1.5) > classAbundance[i - 1]))
			{
				sprintf(logstr, "binNumberReduction: bin %d removed. Abundance %0.4f too close to bin %d [%0.4f]\n", i, classAbundance[i], i - 1, classAbundance[i - 1]);
				writelog(logstr, DEBUG_MUST);
				remove_indicator[i] = true;
				tmp_classnum--;
				change = true;
			}
		}
	}
	if (change == false)
	{
		classifySeq(exclude, exclude_max, 0);
		j = 0;
		for (i = 0; i < class_num; i++)
		{
			j = j + resultSeqNum[i];
		}
		for (i = 0; i < class_num; i++)
		{
			if (resultSeqNum[i] == 0)
			{
				tmp_classnum--;
				change = true;
			}
			else if (((float)resultSeqNum[i] / (float)j) < min_seq_threshold)
			{
				tmp_classnum--;
				change = true;
			}
		}
		/*
		if (change == false)
		{
			for (i = 0; i < class_num - 1; i++)
			{
				if (((classAbundance[i] > classAbundance[i + 1]) && (resultSeqNum[i] < resultSeqNum[i + 1])) || ((classAbundance[i] < classAbundance[i + 1]) && (resultSeqNum[i] > resultSeqNum[i + 1])))
				{
					tmp_classnum--;
					change = true;
				}
			}
		}
		*/
	}
	if (change == true)
	{
		sprintf(logstr, "binNumberReduction: The bin number is reduced from %d to %d\n", class_num, tmp_classnum);
		writelog(logstr, DEBUG_MUST);
		if (tmp_classnum < 2)
		{
			sprintf(logstr, "binNumberReduction: Set the bin number to 2 because bin number should be >= 2.\n");
			writelog(logstr, DEBUG_MUST);
			tmp_classnum = 2;
		}
	}
	else
	{
		sprintf(logstr, "binNumberReduction: No reduction is done for current settings.\n");
		writelog(logstr, DEBUG_MUST);
	}
	writelog("\n", DEBUG_MUST);
	free(remove_indicator);
	return tmp_classnum;
}

void profileManager::writelog(char *ch, int debug_type)
{
	if (debug_type == DEBUG_MUST || (debug_type == DEBUG_NORMAL && is_verbose == true))
	{
		log->write(ch, strlen(ch));
		printf("%s", ch);
	}
}

void profileManager::writelog(float f, int debug_type)
{
	if (debug_type == DEBUG_MUST || (debug_type == DEBUG_NORMAL && is_verbose == true))
	{
		sprintf(str, "%1.2f", f);
		log->write(str, strlen(str));
		printf("%s", str);
	}
}
void profileManager::writelog(double f, int debug_type)
{
	if (debug_type == DEBUG_MUST || (debug_type == DEBUG_NORMAL && is_verbose == true))
	{
		sprintf(str, "%1.2f", f);
		log->write(str, strlen(str));
		printf("%s", str);
	}
}
void profileManager::writelog(int i, int debug_type)
{
	if (debug_type == DEBUG_MUST || (debug_type == DEBUG_NORMAL && is_verbose == true))
	{
		sprintf(str, "%d", i);
		log->write(str, strlen(str));
		printf("%s", str);
	}
}
void profileManager::writelog(unsigned _int64 i, int debug_type)
{
	if (debug_type == DEBUG_MUST || (debug_type == DEBUG_NORMAL && is_verbose == true))
	{
		sprintf(str, "%lu", i);
		log->write(str, strlen(str));
		printf("%s", str);
	}
}

void profileManager::resetRecurMark()
{
	seqNow = seqStart;
	while (seqNow != NULL)
	{
		seqNow->recur_mark = 0;
		seqNow = seqNow->next;
	}
}

void profileManager::revSeq(char *seq, char *result)
{
	int i, len;
	char ch;
	len = strlen(seq);
	for (i = 0; i < len; i++)
	{
		switch (*(seq + i))
		{
			case 'A':
			case 'a':
				ch = 'T';
			break;

			case 'T':
			case 't':
				ch = 'A';
			break;

			case 'C':
			case 'c':
				ch = 'G';
			break;

			case 'G':
			case 'g':
				ch = 'C';
			break;

			default:
				ch = *(seq + i);
		}
		*(result + len - i - 1) = ch;
	}
}

void profileManager::logElapsedTime()
{
	int day, hour, min, sec;
	day = 0;
	hour = 0;
	min = 0;
	sec = (int)difftime(end_time, start_time);
	if (sec > 60)
	{
		min = sec / 60;
		sec = sec % 60;
	}
	if (min > 60)
	{
		hour = min / 60;
		min = min % 60;
	}
	if (hour > 24)
	{
		day = hour / 24;
		hour = hour % 24;
	}
	sprintf(logstr, "%02d days %02d:%02d:%02d\n\n", day, hour, min, sec);
	writelog(logstr, DEBUG_MUST);
}

void profileManager::setIsLength(bool input)
{
	is_len = input;
}

void profileManager::setCoverageThreshold(float input)
{
	coverage_threshold = input;
}

bool profileManager::isFasta()
{
	return is_Fasta;
}

float profileManager::getLikelihood()
{
	return total_likelihood;
}

int profileManager::performTrial(int start_class_num, int stop_class_num, int exclude, int exclude_max)
{
	int i;
	float fTemp;

	fTemp = VERY_SMALL_VALUE;
	writelog("trial: Begin trial process...\n", DEBUG_MUST);
	for (i = start_class_num; i <= stop_class_num; i++)
	{
		initializeEM(i);
		performEM(exclude, exclude_max);
		sprintf(logstr, "Trial: likelihood for bin number %d is %f\n", i, total_likelihood);
		writelog(logstr, DEBUG_MUST);
		if (fTemp < total_likelihood)
		{
			fTemp = total_likelihood;
			trial_class_num = i;
			sprintf(logstr, "trial: Update trial bin number to %d\n", i);
			writelog(logstr, DEBUG_MUST);
		}
		writelog("\n", DEBUG_MUST);
	}

	// When the bin number is decided, perform the EM again to classify sequences.
	sprintf(logstr, "trial: End trial process...\ntrial: The maximized likelihood %0.4f was obtained when bin number is %d\ntrial: Now classifying sequences...\n", fTemp, trial_class_num);
	writelog(logstr, DEBUG_MUST);
	initializeEM(trial_class_num);
	performEM(exclude, exclude_max);
	writelog("trial: Done sequence classification\n", DEBUG_MUST);

	return(trial_class_num);
}

int profileManager::recursiveClassNum(int exclude, int exclude_max)
{
	myStack *stack, *resultStack;
	int return_class_num;
	int i, j, k, itr, tmpnum;
	unsigned short current_part, accumu_part;
	double tempAbundance, tempSize;
	bool need_classify;

	//fstream *fsBin;
	//sprintf(str, "%s.bin_feature", logfile);
	//fsBin = new fstream(str, ios::out);

	stack = new myStack();
	resultStack = new myStack();
	writelog("recursive: Begin recursive process...\n", DEBUG_MUST);

	stack->push((unsigned short)0, (float)0, (float)0);
	accumu_part = 1;
	return_class_num = 0;
	itr = 1;
	j = 0;

	//tmpnum = seqnum;
	while (stack->isContent() == true)
	{
		sprintf(logstr, "\n\nrecursive: iteration %d...\n", itr);
		writelog(logstr, DEBUG_MUST);
		stack->pop(&current_part, &tempAbundance, &tempSize);
		if (current_part == 0)
		{
			writeKmerCount();
		}
		else
		{
			// Scan the sequences to find out the profiles that should be included in current test
			parseFasta(inputFasta, TYPE_RECUR, exclude, exclude_max, current_part);
		}
		initializeEM(2);
		performEM(exclude, exclude_max);
		classifySeq(exclude, exclude_max, current_part);
		// See if this part needs further classification
		need_classify = true;
		for (i = 0; i < 2; i++)
		{
			if (classSize[i] < size_threshold)
			{
				need_classify = false;
			}
		}
		tmpnum = resultSeqNum[0] + resultSeqNum[1];
		if (((classAbundance[1] < classAbundance[0]) && ((classAbundance[1] * 1.5) > classAbundance[0])) || ((classAbundance[0] < classAbundance[1]) && ((classAbundance[0] * 1.5) > classAbundance[1])))
		{
			sprintf(logstr, "recursive: bin abundances %0.4f and %0.4f is too close\n", classAbundance[0], classAbundance[1]);
			writelog(logstr, DEBUG_MUST);
			need_classify = false;
		}
		else if (resultSeqNum[0] == 0 || resultSeqNum[1] == 0)
		{
			sprintf(logstr, "recursive: one of the bins is empty.\n");
			writelog(logstr, DEBUG_MUST);
			need_classify = false;
		}
		/*
		else if (((classAbundance[0] > classAbundance[1]) && (resultSeqNum[0] < resultSeqNum[1])) || ((classAbundance[0] < classAbundance[1]) && (resultSeqNum[0] > resultSeqNum[1])))
		{
			need_classify = false;
		}
		*/
		else if (((float)resultSeqNum[0] / (float)tmpnum) < min_seq_threshold)
		{
			sprintf(logstr, "recursive: sequence number %d is too small compared to total number %d of current run.\n", resultSeqNum[0], tmpnum);
			writelog(logstr, DEBUG_MUST);
			need_classify = false;
		}
		else if (((float)resultSeqNum[1] / (float)tmpnum) < min_seq_threshold)
		{
			sprintf(logstr, "recursive: sequence number %d is too small compared to total number %d of current run.\n", resultSeqNum[1], tmpnum);
			writelog(logstr, DEBUG_MUST);
			need_classify = false;
		}
		if (need_classify == true)
		{
			writelog("recursive: This bin qualifies to further classification. Divide apart.\n", DEBUG_MUST);
			// assign new part number to seqs and corresponding profiles
			seqNow = seqStart;
			while (seqNow != NULL)
			{
				if (seqNow->recur_mark == current_part)
				{
					if (seqNow->class_destination == 0)
					{
						seqNow->recur_mark = accumu_part;
					}
					else if (seqNow->class_destination == 1)
					{
						seqNow->recur_mark = accumu_part + 1;
					}
				}
				seqNow = seqNow->next;
			}
			stack->push(accumu_part + 1, classAbundance[1], classSize[1]);
			stack->push(accumu_part, classAbundance[0], classSize[0]);
			accumu_part = accumu_part + 2;
		}
		else
		{
			writelog("recursive: This bin cannot be classified. Merge the bin into one.\n", DEBUG_MUST);

			// Write abundance and size information
			j++;
			//sprintf(str, "Bin %d\nGenome Abundance: %0.2f\nGenome Size: %0.2f\n\n", j, tempAbundance, tempSize);
			//fsBin->write(str, strlen(str));
			resultStack->push(j, tempAbundance, tempSize);

			seqNow = seqStart;
			while (seqNow != NULL)
			{
				if (seqNow->recur_mark == current_part)
				{
					seqNow->class_destination = return_class_num;
				}
				seqNow = seqNow->next;
			}
			return_class_num++;
		}
		itr++;
	}

	// Reinstate classAbundance and classSize class
	if (classSize != NULL)
	{
		free(classSize);
	}
	if (classAbundance != NULL)
	{
		free(classAbundance);
	}

	sprintf(logstr, "recursive: \nCheck if any resulting classes need to be merged\n");
	writelog(logstr, DEBUG_MUST);

	// Initialize the poisson distribution parameters
	classSize = (double*)mem_handler->allocate(return_class_num * sizeof(double));
	memset(classSize, '\0', return_class_num * sizeof(double));
	classAbundance = (double*)mem_handler->allocate(return_class_num * sizeof(double));
	memset(classAbundance, '\0', return_class_num * sizeof(double));

	for (i = 0; i < return_class_num; i++)
	{
		resultStack->pop(&current_part, &classAbundance[i], &classSize[i]);
	}

	// Merge all classes with similar class abundance values
	k = return_class_num;
	for (i = 0; i < return_class_num; i++)
	{
		for (j = i + 1; j < return_class_num; j++)
		{
			if (((classAbundance[j] < classAbundance[i]) && ((classAbundance[j] * 1.5) > classAbundance[i])) || ((classAbundance[i] < classAbundance[j]) && ((classAbundance[i] * 1.5) > classAbundance[j])))
			{
				k--;
			}
		}
	}

	if (k != return_class_num)
	{
		sprintf(logstr, "recursive: \nClass number being merged into %d. Perform final classification\n", k);
		writelog(logstr, DEBUG_MUST);

		return_class_num = k;
		resetRecurMark();
		parseFasta(inputFasta, TYPE_RECUR, exclude, exclude_max, 0);
		initializeEM(return_class_num);
		performEM(exclude, exclude_max);
		classifySeq(exclude, exclude_max, 0);
		writeBinFeature();
	}

	sprintf(logstr, "recursive: \nThe sequences can be divided into %d classes using recursive method.\n", return_class_num);
	writelog(logstr, DEBUG_MUST);

	free(stack);
	free(resultStack);
	class_num = return_class_num;

	// Write total number of bins and close the feature file
	//sprintf(str, "Total number of bins: %d\n\n", class_num);
	//fsBin->write(str, strlen(str));
	//fsBin->close();
	//delete(fsBin);

	return return_class_num;
}
